package frc.robot.subsystems;

import java.util.function.DoubleSupplier;

import com.ctre.phoenix6.hardware.CANcoder;
import com.revrobotics.spark.SparkMax;
import com.revrobotics.spark.SparkBase.PersistMode;
import com.revrobotics.spark.SparkBase.ResetMode;
import com.revrobotics.spark.SparkLowLevel.MotorType;

import edu.wpi.first.math.controller.ArmFeedforward;
import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.math.util.Units;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import frc.robot.constants.ArmConstants;

public class Arm extends SubsystemBase {
    protected SparkMax armMotor;

    private CANcoder canCoder;

    private PIDController positionController;
    private PIDController velocityController;

    private ArmFeedforward feedForward;

    public Arm() {
        armMotor = new SparkMax(
            ArmConstants.kArmMotorID, 
            MotorType.kBrushless
        );

        armMotor.configure(ArmConstants.motorConfig, ResetMode.kResetSafeParameters, PersistMode.kPersistParameters);

        positionController = new PIDController(
            ArmConstants.kPositionalP, 
            ArmConstants.kPositionalI, 
            ArmConstants.kPositionalD
        );
        
        // TODO: Generate constants for continuous input range based on CANcoder configuration?
        positionController.enableContinuousInput(Units.degreesToRadians(-180), Units.degreesToRadians(179));
        positionController.setTolerance(ArmConstants.kPositionalTolerance);

        velocityController = new PIDController(
            ArmConstants.kVelocityP, 
            ArmConstants.kVelocityI, 
            ArmConstants.kVelocityD
        );

        velocityController.setTolerance(ArmConstants.kVelocityTolerance);

        canCoder = new CANcoder(ArmConstants.kCANcoderID);
        canCoder.getConfigurator().apply(ArmConstants.canCoderConfig);
    }

    /**
     * Returns whether or not the motion is safe relative to the encoder's current position
     * and the arm safe stow position
     * 
     * @return Is the motion safe
     */
    public boolean isMotionSafe() {
        return isMotionSafe(getEncoderPosition());
    }

    /**
     * Returns whether or not the motion is safe relative to some target position and the
     * arm safe stow position
     * 
     * @param motionTarget The target position to determine the safety of
     * @return Is the motion safe
     */
    public boolean isMotionSafe(double motionTarget) {
        return motionTarget > ArmConstants.kArmSafeStowPosition;
    }

    //manual command that keeps ouput speed consistent no matter the direction
    public Command runArm(DoubleSupplier speed) {
        return run(() -> {
            double realSpeedTarget = speed.getAsDouble() * ArmConstants.kArmMaxVelocity;

            double voltsOut = velocityController.calculate(
                getEncoderVelocity(), 
                realSpeedTarget
            ) + feedForward.calculate(
                getEncoderPosition(), 
                getEncoderVelocity()
            );

            armMotor.setVoltage(voltsOut);
        });
    }

    public Command goToSetpoint(double setpoint, double timeout) {
        return run(() -> {
            double voltsOut = positionController.calculate(
                getEncoderPosition(), 
                setpoint
            ) + feedForward.calculate(
                getEncoderPosition(), 
                getEncoderVelocity()
            );

            armMotor.setVoltage(voltsOut);
        }).until(positionController::atSetpoint).withTimeout(timeout);
    }

    public double getEncoderPosition() {
        return Units.rotationsToRadians(canCoder.getAbsolutePosition().getValueAsDouble());
    }

    public double getEncoderVelocity() {
        return Units.rotationsToRadians(canCoder.getVelocity().getValueAsDouble());
    }
}
