package frc.robot.subsystems;

import java.util.function.DoubleSupplier;

import com.revrobotics.RelativeEncoder;
import com.revrobotics.spark.SparkMax;
import com.revrobotics.spark.SparkBase.PersistMode;
import com.revrobotics.spark.SparkBase.ResetMode;
import com.revrobotics.spark.SparkLowLevel.MotorType;

import edu.wpi.first.math.controller.ElevatorFeedforward;
import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.wpilibj.DigitalInput;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import frc.robot.constants.ElevatorConstants;

public class Elevator extends SubsystemBase {
    protected SparkMax elevatorMotor1;
    protected SparkMax elevatorMotor2;

    protected RelativeEncoder encoder;

    private DigitalInput bottomLimitSwitch;

    private PIDController positionController;
    private PIDController velocityController;

    private ElevatorFeedforward feedForward;
    
    public Elevator() {
        elevatorMotor1 = new SparkMax(
            ElevatorConstants.kElevatorMotor1ID, 
            MotorType.kBrushless
        );
        
        elevatorMotor2 = new SparkMax(
            ElevatorConstants.kElevatorMotor2ID,
            MotorType.kBrushless
        );

        elevatorMotor1.configure(
            ElevatorConstants.motorConfig, 
            ResetMode.kResetSafeParameters, 
            PersistMode.kPersistParameters
        );

        elevatorMotor2.configure(
            ElevatorConstants.motorConfig.follow(ElevatorConstants.kElevatorMotor1ID), 
            ResetMode.kResetSafeParameters, 
            PersistMode.kPersistParameters
        );

        encoder = elevatorMotor1.getEncoder();

        bottomLimitSwitch = new DigitalInput(
            ElevatorConstants.kBottomLimitSwitchID
        );

        positionController = new PIDController(
            ElevatorConstants.kPositionControllerP, 
            ElevatorConstants.kPositionControllerI, 
            ElevatorConstants.kPositionControllerD
        );

        velocityController = new PIDController(
            ElevatorConstants.kVelocityControllerP, 
            ElevatorConstants.kVelocityControllerI, 
            ElevatorConstants.kVelocityControllerD
        );

        feedForward = new ElevatorFeedforward(
            ElevatorConstants.kFeedForwardS, 
            ElevatorConstants.kFeedForwardG, 
            ElevatorConstants.kFeedForwardV
        );
    }

    /**
     * Returns whether or not the motion is safe relative to the encoder's current position
     * and the elevator brace position
     * 
     * @return Is the motion safe
     */
    public boolean isMotionSafe() {
        return isMotionSafe(getEncoderPosition());
    }

    /**
     * Returns whether or not the motion is safe relative to some target position and the elevator
     * brace position
     * 
     * @param motionTarget The target position to determine the safety of
     * @return Is the motion safe
     */
    public boolean isMotionSafe(double motionTarget) {
        return motionTarget > ElevatorConstants.kElevatorBracePosition;
    }

    /**
     * A manual translation command that will move the elevator using a consistent velocity disregarding direction
     * 
     * @param speed How fast the elevator moves
     * @return Sets motor voltage to move the elevator relative to the speed parameter
     */
    public Command runAssistedElevator(DoubleSupplier speed) {
        return run(() -> {
            double realSpeedTarget = speed.getAsDouble() * ElevatorConstants.kElevatorMaxVelocity;

            double voltsOut = velocityController.calculate(
                encoder.getVelocity(), 
                realSpeedTarget
            ) + feedForward.calculate(realSpeedTarget);

            elevatorMotor1.setVoltage(voltsOut);
        }).until(
            () -> bottomLimitSwitch.get() || encoder.getPosition() >= ElevatorConstants.kElevatorMaxHeight);
    }

    /**
     * A manual translation command that uses feed forward calculation to maintain position
     * 
     * @param speed The speed at which the elevator translates
     * @return Sets motor voltage to translate the elevator and maintain position
     */
    public Command runManualElevator(double speed) {
        return run(() -> {
            elevatorMotor1.set(speed);
        });
    }
    
    /**
     * Moves the elevator to a target destination (setpoint)
     * 
     * @param setpoint Target destination of the subsystem
     * @param timeout Time to achieve the setpoint before quitting
     * @return Sets motor voltage to achieve the target destination
     */
    public Command goToSetpoint(double setpoint, double timeout) {
        return run(() -> {
            double voltsOut = positionController.calculate(
                encoder.getPosition(),
                setpoint
            ) + feedForward.calculate(0);

            elevatorMotor1.setVoltage(voltsOut);
        }).until(
            () -> positionController.atSetpoint() || bottomLimitSwitch.get()
        ).withTimeout(timeout);
    }

    /**
     * Returns the current encoder position
     * 
     * @return Current encoder position
     */
    public double getEncoderPosition() {
        return encoder.getPosition();
    }

    /**
     * Returns the value of the bottom limit switch on the elevator (false = disabled, true = enabled)
     * 
     * @return The value of bottomLimitSwitch
     */
    public boolean getBottomLimitSwitch() {
        return bottomLimitSwitch.get();
    }
}